(function (ng) {

    /**
     * @ngdoc type
     * @module apeman-ui-contrib-angular-decorating
     * @name ApDecorator
     * @description Object property decorator.
     */
    "use strict";

    ng
        .module('apeman-ui-contrib-angular-decorating')
        .factory('ApDecorator', function defineApDecorator(apDecorationHistory) {
            function ApDecorator(config) {
                var s = this;
                config = config || {};
                var decoratorId = config.decoratorId || String(Object.keys(ApDecorator.decoratorIds).length);
                ApDecorator.decoratorIds[decoratorId] = true;
                s.decoratorId = decoratorId;
            }

            ApDecorator.decoratorIds = {};

            ApDecorator.prototype = {
                /**
                 * @ngdoc method
                 * @name ApDecorator#decorate
                 * @description Decorate an object functions.
                 * @param {object} target - Target object
                 * @param {string} key - Name of property to decorate
                 * @param {function} factory - Factory function to create decorated property.
                 */
                decorate: function (target, key, factory) {
                    var s = this,
                        decoratorId = s.decoratorId;

                    var origin = target[key];
                    target[key] = factory(origin);

                    apDecorationHistory.pushHistory(decoratorId, target, key, origin);

                    return s;
                },
                /**
                 * @ngdoc method
                 * @name ApDecorator#popDecoration
                 * @description Undo last decoration.
                 * @param {object} target - Target object
                 * @param {string} key - Property key to pop.
                 */
                undoDecoration: function (target, key) {
                    var s = this,
                        decoratorId = s.decoratorId;

                    var origin = apDecorationHistory.popHistory(decoratorId, target, key);
                    if (!origin) {
                        return;
                    }
                    var decoration = target[key];
                    target[key] = origin;
                    return decoration;
                },
                /**
                 * @ngdoc method
                 * @name ApDecorator#popDecoration
                 * @description Undo last decoration.
                 * @param {object} target - Target object
                 * @param {string} key - Property key to pop.
                 */
                undoAllDecorations: function (target, key) {
                    var s = this,
                        decorations = [];
                    var decoration;
                    while (true) {
                        decoration = s.undoDecoration(target, key);
                        if (decoration) {
                            decorations.push(decoration)
                        } else {
                            break;
                        }
                    }
                    return decorations;
                }
            };

            return ApDecorator;
        });


})(angular);