(function (ng) {

    /**
     * @ngdoc service
     * @module apeman-ui-contrib-angular-decorating
     * @name apDecorationHistory
     * @description Manage decoration history.
     */

    "use strict";

    ng
        .module('apeman-ui-contrib-angular-decorating')
        .service('apDecorationHistory', function pDecorationHistory() {
            var s = this;
            /**
             * @ngdoc method
             * @name apDecorationHistory#pushHistory
             * @description Push a history.
             * @param {string} decoratorId - Id of decorator.
             * @param {object} target - Target object.
             * @param {string} key - Property key.
             * @param {*} val - Property value.
             */
            s.pushHistory = function (decoratorId, target, key, val) {
                s.histories(decoratorId, target, key).push(val);
            };

            /**
             * @ngdoc method
             * @name apDecorationHistory#pushHistory
             * @description pop a history.
             * @param {string} decoratorId - Id of decorator.
             * @param {object} target - Target object.
             * @param {string} key - Property key.
             * @returns {*} - Popped history.
             */
            s.popHistory = function (decoratorId, target, key) {
                return s.histories(decoratorId, target, key).pop();
            };

            s.histories = function (decoratorId, target, key) {
                var fullKey = [decoratorId, key].join('.');
                target['$apDecorationHistory'] = target['$apDecorationHistory'] || {};
                target['$apDecorationHistory'][fullKey] = target['$apDecorationHistory'][fullKey] || [];
                return target['$apDecorationHistory'][fullKey];
            }
        })

})(angular);