(function (ng) {
    /**
     * @ngdoc module
     * @module apeman-ui-contrib-angular-decorating
     * @name apeman-ui-contrib-angular-decorating
     * @description
     * apeman ui module for angular to decorate functions
     */

    "use strict";

    ng
        .module('apeman-ui-contrib-angular-decorating', [
        ]);
})(angular);(function (ng) {

    /**
     * @ngdoc type
     * @module apeman-ui-contrib-angular-decorating
     * @name ApDecorator
     * @description Object property decorator.
     */
    "use strict";

    ng
        .module('apeman-ui-contrib-angular-decorating')
        .factory('ApDecorator', function defineApDecorator(apDecorationHistory) {
            function ApDecorator(config) {
                var s = this;
                config = config || {};
                var decoratorId = config.decoratorId || String(Object.keys(ApDecorator.decoratorIds).length);
                ApDecorator.decoratorIds[decoratorId] = true;
                s.decoratorId = decoratorId;
            }

            ApDecorator.decoratorIds = {};

            ApDecorator.prototype = {
                /**
                 * @ngdoc method
                 * @name ApDecorator#decorate
                 * @description Decorate an object functions.
                 * @param {object} target - Target object
                 * @param {string} key - Name of property to decorate
                 * @param {function} factory - Factory function to create decorated property.
                 */
                decorate: function (target, key, factory) {
                    var s = this,
                        decoratorId = s.decoratorId;

                    var origin = target[key];
                    target[key] = factory(origin);

                    apDecorationHistory.pushHistory(decoratorId, target, key, origin);

                    return s;
                },
                /**
                 * @ngdoc method
                 * @name ApDecorator#popDecoration
                 * @description Undo last decoration.
                 * @param {object} target - Target object
                 * @param {string} key - Property key to pop.
                 */
                undoDecoration: function (target, key) {
                    var s = this,
                        decoratorId = s.decoratorId;

                    var origin = apDecorationHistory.popHistory(decoratorId, target, key);
                    if (!origin) {
                        return;
                    }
                    var decoration = target[key];
                    target[key] = origin;
                    return decoration;
                },
                /**
                 * @ngdoc method
                 * @name ApDecorator#popDecoration
                 * @description Undo last decoration.
                 * @param {object} target - Target object
                 * @param {string} key - Property key to pop.
                 */
                undoAllDecorations: function (target, key) {
                    var s = this,
                        decorations = [];
                    var decoration;
                    while (true) {
                        decoration = s.undoDecoration(target, key);
                        if (decoration) {
                            decorations.push(decoration)
                        } else {
                            break;
                        }
                    }
                    return decorations;
                }
            };

            return ApDecorator;
        });


})(angular);(function (ng) {
    /**
     * @ngdoc service
     * @module apeman-ui-contrib-angular-decorating
     * @name apDecorate
     * @description Function decorator. This is an instance of apeman-ui-contrib-angular-decorating.ApDecorator
     */

    "use strict";

    ng
        .module('apeman-ui-contrib-angular-decorating')
        .service('apDecorate', function ApDecorate(ApDecorator) {
            var s = this;
            s.__proto__ = new ApDecorator();


            /**
             * @ngdoc method
             * @name apDecorate#create
             * @module apeman-ui-contrib-angular-decorating
             * @description Create a new context.
             */
            s.create = function () {
                return new ApDecorator();
            };

        });


})(angular);(function (ng) {

    /**
     * @ngdoc service
     * @module apeman-ui-contrib-angular-decorating
     * @name apDecorationHistory
     * @description Manage decoration history.
     */

    "use strict";

    ng
        .module('apeman-ui-contrib-angular-decorating')
        .service('apDecorationHistory', function pDecorationHistory() {
            var s = this;
            /**
             * @ngdoc method
             * @name apDecorationHistory#pushHistory
             * @description Push a history.
             * @param {string} decoratorId - Id of decorator.
             * @param {object} target - Target object.
             * @param {string} key - Property key.
             * @param {*} val - Property value.
             */
            s.pushHistory = function (decoratorId, target, key, val) {
                s.histories(decoratorId, target, key).push(val);
            };

            /**
             * @ngdoc method
             * @name apDecorationHistory#pushHistory
             * @description pop a history.
             * @param {string} decoratorId - Id of decorator.
             * @param {object} target - Target object.
             * @param {string} key - Property key.
             * @returns {*} - Popped history.
             */
            s.popHistory = function (decoratorId, target, key) {
                return s.histories(decoratorId, target, key).pop();
            };

            s.histories = function (decoratorId, target, key) {
                var fullKey = [decoratorId, key].join('.');
                target['$apDecorationHistory'] = target['$apDecorationHistory'] || {};
                target['$apDecorationHistory'][fullKey] = target['$apDecorationHistory'][fullKey] || [];
                return target['$apDecorationHistory'][fullKey];
            }
        })

})(angular);